/*-
 * Copyright (C) @BABOLO  2008 Feb 10
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2008 Feb 10\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: db_pgsql.c,v 1.3 2008/02/15 23:31:22 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <errno.h>
#include <mife.h>
#include <err.h>
#include <libpq-fe.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

#define PRINTNOTNULL(A) (!(A) ? "" : (A))
#define PRINTANDNULL(A) PRINTNOTNULL(A), ((A) ? "~" : "")
#define PQSTATNDNULL(A) PRINTANDNULL(PQparameterStatus(exenv->options->conn[r].conn, (A)))
#define ALIGN8(A)  popt[i].A, (strlen(popt[i].A) < 8 ? "		" : "	")
#define ALIGN16(A) popt[i].A                                                                          \
      , ((strlen(popt[i].A) < 8) ? "		" : (strlen(popt[i].envvar) < 16) ? "	" : " ")

static void
NoticeProcessor(void *options, const char *message) {
    ifBLIN_QV1(((pgoblin_main *)options)->flags) fprintf(stderr, "%s", message);
}

int
pgoblin_shurecon(pgoblin_exenv *exenv, int r) {
    PQconninfoOption *popt;
    int i, ex = EX_OK;
    PGresult *res;
#   define DBSET (pgoblin.dbasetable[exenv->options->conn[r].flags & PGOBLIN_DB_TYPE])

    if  (  exenv->options->conn[r].conn != NULL
        && PQstatus(exenv->options->conn[r].conn) == CONNECTION_OK
        ) return(0);
    exenv->options->conn[r].flags &= ~PGOBLIN_DB_TYPE;
    exenv->options->conn[r].flags |= (PGOBLIN_DB_PGSQL & PGOBLIN_DB_TYPE);
    ifBLIN_QV5(exenv->options->flags) {
        fprintf( stderr
               , "Connection defaults:\n"
                 "size=chr keyword	envvar		compiled val		label\n"
               );
        popt = PQconndefaults();
        for (i = 0; popt[i].keyword; i++) {
            fprintf( stderr
                   , "%4d=%s~	%s%s%s%s%s%s%s%s%s\n"
                   , popt[i].dispsize
                   , popt[i].dispchar
                   , ALIGN8(keyword)
                   , ALIGN16(envvar)
                   , PRINTNOTNULL(popt[i].compiled)
                   , !(popt[i].compiled) ? "	" : strlen(popt[i].compiled) < 7 ? "~	" : "~ "
                   , PRINTNOTNULL(popt[i].val)
                   , !(popt[i].val)
                   ? "		"
                   : strlen(popt[i].val) < 7 ? "~		" : "~	"
                   , popt[i].label
                   );
        }
        PQconninfoFree(popt);
            fprintf( stderr
                   , "PGREALM=%s%s PGDATESTYLE=%s%s PGTZ=%s%s PGCLIENTENCODING=%s%s PGGEQO=%s%s\n"
                   , PRINTANDNULL(getenv("PGREALM"))
                   , PRINTANDNULL(getenv("PGDATESTYLE"))
                   , PRINTANDNULL(getenv("PGTZ"))
                   , PRINTANDNULL(getenv("PGCLIENTENCODING"))
                   , PRINTANDNULL(getenv("PGGEQO"))
                   );
    }
    exenv->options->conn[r].conn = PQsetdbLogin( exenv->options->conn[r].host
                                               , exenv->options->conn[r].port
                                               , NULL, NULL
                                               , exenv->options->conn[r].dbname
                                               , exenv->options->conn[r].username
                                               , NULL
                                               );
    exenv->options->conn[r].flags |= PGOBLIN_CLOSECONN;
    ifBLIN_QV5(exenv->options->flags) {
        fprintf(stderr, "PQoptions=%s~\n", PQoptions(exenv->options->conn[r].conn));
        fprintf( stderr
               , "PQdb=%s~ PQuser=%s~ PQpass=%s%s "
                 "PQhost=%s%s PQport=%s%s "
                 "PQtty=%s%s\n"
               , PQdb(exenv->options->conn[r].conn), PQuser(exenv->options->conn[r].conn)
               , PRINTANDNULL(PQpass(exenv->options->conn[r].conn))
               , PRINTANDNULL(PQhost(exenv->options->conn[r].conn))
               , PRINTANDNULL(PQport(exenv->options->conn[r].conn))
               , PRINTANDNULL(PQtty(exenv->options->conn[r].conn))
               );
        fprintf( stderr
               , "server_version=%s%s "
                 "client_encoding=%s%s "
                 "is_superuser=%s%s "
                 "session_authorization=%s%s "
                 "DateStyle=%s%s\n"
               , PQSTATNDNULL("server_version")
               , PQSTATNDNULL("client_encoding")
               , PQSTATNDNULL("is_superuser")
               , PQSTATNDNULL("session_authorization")
               , PQSTATNDNULL("DateStyle")
               );
        fprintf( stderr
               , "PQprotocolVersion=%d PQsocket=%d PQbackendPID=%d\n"
               , PQprotocolVersion(exenv->options->conn[r].conn)
               , PQsocket(exenv->options->conn[r].conn)
               , PQbackendPID(exenv->options->conn[r].conn)
               );
    }
    ifBLIN_QV5(exenv->options->flags)
        fprintf(stderr, "PQstatus %d\n", PQstatus(exenv->options->conn[r].conn));
    PQsetNoticeProcessor(exenv->options->conn[r].conn, NoticeProcessor, exenv->options);
    if  (PQstatus(exenv->options->conn[r].conn) != CONNECTION_BAD) {
        if  (  exenv->options->conn[r].flags & PGOBLIN_NEEDINTIME
            && !(exenv->options->conn[r].flags & (PGOBLIN_INT64TIME | PGOBLIN_FLOATTIME))
            ) {
            if  (!exenv->options->conn[r].intran) {
/*          if  (PQtransactionStatus(exenv->options->conn[r].conn) != PQTRANS_INTRANS) { */
                PQEXEC(res, "BEGIN", PGRES_COMMAND_OK, -1, exenv->options->conn[r].conn);
                PQclear(res);
            }
            PQEXEC( res
                  , "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR "
                    "SELECT '2000-01-01 00:00:01'::timestamp;"
                    "FETCH 1 FROM pgoblin"
                  , PGRES_TUPLES_OK, -1
                  , exenv->options->conn[r].conn
                  );
            if  (  PQgetisnull(res, 0, 0)
                || (PQgetlength(res, 0, 0) < 8)
                ) {
                ex = EX_DATAERR;
                errno = EINVAL;
            } else {
                if  (*((u_char*)(PQgetvalue(res, 0, 0))) == 077) {
                    exenv->options->conn[r].flags |= PGOBLIN_FLOATTIME;
                    ifBLIN_QV3(exenv->options->flags)
                        fprintf(stderr, "conn %d float timestamp format\n", r);
                } else if (*((u_char*)(PQgetvalue(res, 0, 0))) == 0) {
                    exenv->options->conn[r].flags |= PGOBLIN_INT64TIME;
                    ifBLIN_QV3(exenv->options->flags)
                        fprintf(stderr, "conn %d int64 timestamp format\n", r);
                } else {
                    ifBLIN_QV1(exenv->options->flags)
                        fprintf(stderr, "conn %d unknown timestamp format\n", r);
                    ex = EX_DATAERR;
                    errno = EFTYPE;
            }   }
            PQclear(res);
            PQEXEC(res, "CLOSE pgoblin", PGRES_COMMAND_OK, -1, exenv->options->conn[r].conn);
            PQclear(res);
            if  (!exenv->options->conn[r].intran) {
                PQEXEC(res, "END", PGRES_COMMAND_OK, -1, exenv->options->conn[r].conn);
                PQclear(res);
        }   }
        return(0);
    }
    ifBLIN_QV1(exenv->options->flags)
        warnx("Connection to PostgreSQL failed: %s\n", PQerrorMessage(exenv->options->conn[r].conn));
    return(1);
#   undef DBSET
}

pgoblin_dbases pgoblin_db_pgsql =
{ "pgsql"
, pgoblin_shurecon
, (void*(*)(void*, const char *))PQexec
, (int(*)(void*))PQntuples
, (int(*)(void*))PQnfields
, (int(*)(void*, int, int))PQgetisnull
, (int(*)(void*, int, int))PQgetlength
, (char*(*)(void*, int, int))PQgetvalue
, (void(*)(void*))PQclear
, (void(*)(void*))PQfinish
, (void(*)(void*))PQfreemem
};
