/*-
 * Copyright (C) @BABOLO  2002 Dec 23
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 23\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: parser.c,v 1.30 2008/02/14 01:06:43 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <mife.h>
#include <err.h>
#include <babolo/parser.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"
#include "cmds.h"

#define L0 0 /* \0           */
#define Ll 1 /*   */
#define Lb 2 /*        */
#define Ls 3 /*    */
#define Ld 4 /* #            */
#define Lo 5 /*     */
#define Lx 6

static const u_char *clasnames[] =
{ "L0", "Ll", "Lb", "Ls", "Ld", "Lo", "Lx"};

static const char class[256] =
{ L0, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lb, Ll, Ll,  Lo, Ll, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lb, Lo, Lo, Ld,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, -1, -2, -3, -4,  -5, -6, -7, -8,  -9,-10, Lo, Lo,  Lo, Lo, Lo, Lo

, Lo,-11,-12,-13, -14,-15,-16,-17, -18,-19,-20,-21, -22,-23,-24,-25
,-26,-27,-28,-29, -30,-31,-32,-33, -34,-35,-36, Lo,  Lo, Lo, Lo,-63
, Lo,-37,-38,-39, -40,-41,-42,-43, -44,-45,-46,-47, -48,-49,-50,-51
,-52,-53,-54,-55, -56,-57,-58,-59, -60,-61,-62, Lo,  Lo, Lo, Lo, Lo

, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo

, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
};

void
pgoblin_verify(pgoblin_main *options) {
    int i;
    u_int32_t m;
    pgoblin_command cmd;

    if  (pgoblin.symparam != SYMREGS) {
        printf("!!! pgoblin.symparam=%d <> SYMREGS=%d\n", pgoblin.symparam, SYMREGS);
    }
    if  (pgoblin.binparam != BINREGS) {
        printf("!!! pgoblin.binparam=%d <> BINREGS=%d\n", pgoblin.binparam, BINREGS);
    }
    if  (pgoblin.max != PGOBLIN_max) {
        printf("!!! pgoblin.max=%d <> PGOBLIN_max=%d\n", pgoblin.max, PGOBLIN_max);
    }
    if  (pgoblin.binparams[0] != PGOBLIN_ARG1S) {
        printf( "!!! pgoblin.binparams[0]=%d <> PGOBLIN_ARG1S=%d\n"
              , pgoblin.binparams[0], PGOBLIN_ARG1S
              );
    }
    if  (pgoblin.binparams[1] != PGOBLIN_ARG2S) {
        printf( "!!! pgoblin.binparams[1]=%d <> PGOBLIN_ARG2S=%d\n"
              , pgoblin.binparams[1], PGOBLIN_ARG2S
              );
    }
    if  (pgoblin.binparams[2] != PGOBLIN_ARG3S) {
        printf( "!!! pgoblin.binparams[2]=%d <> PGOBLIN_ARG3S=%d\n"
              , pgoblin.binparams[2], PGOBLIN_ARG3S
              );
    }
    if  (pgoblin.binparams[3] != PGOBLIN_ARG4S) {
        printf( "!!! pgoblin.binparams[3]=%d <> PGOBLIN_ARG4S=%d\n"
              , pgoblin.binparams[3], PGOBLIN_ARG4S
              );
    }
    for (i = 1; i < pgoblin.binparam; i++) {
        if  (pgoblin.binparams[i - 1] + PGOBLIN_BPARG != pgoblin.binparams[i]) {
            printf( "!!! pgoblin.binparams[%d]=%d + %d <> pgoblin.binparams[%d]=%d\n"
                  , i - 1, pgoblin.binparams[i - 1]
                  , PGOBLIN_BPARG
                  , i, pgoblin.binparams[i]
                  );
    }   }
    for (i = 0; i < PGOBLIN_REGSIZE; i++) {
        if  ((~class[pgoblin_regn[i] & 0xFF]) != i) {
            printf("reg[%d <> %d] name %c\n", i, ~class[pgoblin_regn[i] & 0xFF], pgoblin_regn[i]);
    }   }
    for (i = 0; i < 256; i++) {
        if  ((class[i] < 0) && (pgoblin_regn[~class[i] & 0xFF] != i)) {
            printf("name %c <> %c reg[%d]\n", pgoblin_regn[~class[i] & 0xFF], i, ~class[i] & 0xFF);
    }   }
    for (cmd = 1; cmd < pgoblin.max; cmd++) {
        m = 0;
        if  (!(~pgoblin.a[cmd].flags & PGOBLIN_ILLEGAL)) {
            if  (pgoblin.a[cmd].car) printf("!executor for illegal command %03X\n", cmd);
            if  (pgoblin.a[cmd].str)
                printf("!name=%s~ for illegal command %03X\n", pgoblin.a[cmd].str, cmd);
        } else {
            if  (!pgoblin.a[cmd].str) {
                printf("!! No name for legal command %03X\n", cmd);
            } else if  (babolo_testword(&cmds, pgoblin.a[cmd].str) != cmd) {
                printf( "!!! cmd=%03X: %s=%03X\n"
                      , cmd, pgoblin.a[cmd].str, babolo_testword(&cmds, pgoblin.a[cmd].str)
                      );
            }
            if  (pgoblin.a[cmd].flags & PGOBLIN_IGNORE) {
                if  (pgoblin.a[cmd].car) printf("!executor for ignored command %03X\n", cmd);
            } else if (pgoblin.a[cmd].flags && !pgoblin.a[cmd].car) {
                printf("!! No executor for legal command %03X\n", cmd);
            }
            if  (pgoblin.a[cmd].flags & PGOBLIN_NOLITER) {
                if  (!(cmd & PGOBLIN_EXTENDED)) printf("!short command %03X without literal\n", cmd);
        }   }
        for (i = 0; i < pgoblin.symparam; i++) {
            if  (pgoblin.a[cmd].flags & pgoblin.p[i].sym) {
                if  (!pgoblin.a[cmd].ch[PREG(cmd, i)]) {
                    printf("!! cmd=%03X: pgoblin.a[cmd].flags <> pgoblin.a[cmd].ch[PREG(%d)]\n", cmd, i);
                    ifBLIN_QV2(options->flags) {
                        printf( "pgoblin.a[%03X].flags=%08X pgoblin.p[%d].sym=%02X, bin=%02X, mask=%02X "
                                "PREG(%03X,%d)=%d pgoblin.a[%03X].ch[%d]=%02X\n"
                              , cmd
                              , pgoblin.a[cmd].flags
                              , i
                              , pgoblin.p[i].sym
                              , pgoblin.p[i].bin & 0xFF
                              , pgoblin.p[i].mask
                              , cmd, i, PREG(cmd, i)
                              , cmd, PREG(cmd, i), pgoblin.a[cmd].ch[PREG(cmd, i)]
                              );
                    }
                } else {
                    if  (m & (1 << PREG(cmd, i))) {
                        printf("!!! Second hand r=%d for cmd=%03X\n", i, cmd);
                    }
                    m |= (1 << PREG(cmd, i));
        }   }   }
        for (i = 0; i < pgoblin.binparam; i++) {
            if  (!(m & (1 << i)) && pgoblin.a[cmd].ch[i]) {
                printf("!!! Unused r=%d for cmd=%03X\n", i, cmd);
}   }   }   }

static const u_char *statenames[] =
{ "bg", "bs", "nl", "nq", "tc", "tb", "tp", "sc", "cl", "cq", "ml", "mc", "tq", "xx"};

enum states
{ bg /*                                */
, bs /*                                 */
, nl /*                           */
, nq /*     ?       */
, tc /*                 */
, tb /*         */
, tp /*                                */
, sc /*       () */
, cl /*                        */
, cq /* ?                       */
, ml /*                           */
, mc /*                                 */
, tq /*   ?                 */
, xx
};
static const u_char *pn[] = 
{ "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "- ", "- ", "Cp", "- ", "Cb", "Ce", "Cc", "Ci"
, "Co", "Cl", "- ", "- ", "Cx", "Cy", "- ", "- "
};

#define Cp  0x200000 /*             */
#define Cb  0x080000 /*       */
#define Ce  0x040000 /*        */
#define Cc  0x020000 /*               */
#define Ci  0x010000 /*     */
#define Co  0x008000 /*             */
#define Cl  0x004000 /*  lexor                   */
#define Cx  0x000800 /*                 */
#define Cy  0x000400 /*                    */
#define C_state 0xFF /*           */
#define Cio (Ci | Co)

static u_int32_t automa[xx][Lx] =
/*    \0 L0        \n Ll      \b Lb   Ls       # Ld       * Lo  *    @          */
{{    Cx|xx,          bs,        bs,        bs,        nq,        bs}/* bg @                   */
,{    Cx|xx,          nl,        bs,        bs,        bs,        bs}/* bs @                   */
,{    Cx|xx,          nl,        bs,        bs,        nq,        bs}/* nl @                */
,{ Co|Cl|xx,  Co|Cl|  cl,  Co|Cl|tb,  Co|Cl|tc,        bs,  Co|Cl|sc}/* nq @ 1           */
,{    Cl|xx,     Cl|  cl,     Cl|tb,     Cl|tc,     Cl|sc,     Cl|sc}/* tc                    */
,{       xx,          cl,        tb, Cp|    tp,        sc,        sc}/* tb        */
,{       xx,          cl,        sc, Cp|    tp,        sc,        sc}/* tp           */
,{       xx,          cl,        sc,        sc,        sc,        sc}/* sc                */
,{Ci |   xx, Cb|   Cc|ml, Cb|Cc| mc, Cb|Cc| mc,        cq, Cb|Cc| mc}/* cl               */
,{Cio|   xx, Cio|     ml, Cio|   tb, Cio|Cl|tc, Cb|Cc| mc, Cio|   sc}/* q   1           */
,{Ci |   xx,    Ce|Cc|ml,    Cc| mc,    Cc| mc,        tq,    Cc| mc}/* ml #             */
,{    Cy|xx,    Ce|Cc|ml,    Cc| mc,    Cc| mc, Cc|    mc,    Cc| mc}/* mc #                 */
,{Cio|Cl|xx, Cio|Cl|  cl, Cio|Cl|tb, Cio|Cl|tc, Cc|    mc, Cio|Cl|sc}/* tq # 1           */
};                                                                      /* #                  */

int
pgoblin_parser(pgoblin_main *options, u_char *text, size_t txlen, pgoblin_prog **ppgm) {
    int ex = EX_OK;
    size_t offset;
    enum states state;
    u_int32_t control, localflags;
    const u_char **qq, *qp;
    u_char s, *r_beg = NULL, *r_end = NULL, *literal = NULL;
    char p;
    u_int32_t r_lex, r_cmd;
    u_int64_t r_parm;
    int rs, lfcount, i;
    pgoblin_prog *pgm = NULL;

    ifBLIN_QV3(options->flags) fprintf( stderr
                                      , "+pgoblin_parser: options %s, ppgm %s txt %d\n"
                                      , options ? "OK" : "NULL"
                                      , ppgm ? "OK" : "NULL"
                                      , txlen
                                      );
    if  (!text) {
        ifBLIN_QV1(options->flags) warnx("No pgoblin program text");
        ERROUT(EX_DATAERR, EINVAL);
    } else if (!txlen) {
        ifBLIN_QV1(options->flags) warnx("Empty pgoblin program text");
        ERROUT(EX_DATAERR, EINVAL);
    } else if (!ppgm) {
        ifBLIN_QV1(options->flags) warnx("No pgoblin program");
        ERROUT(EX_DATAERR, EINVAL);
    } else if ((pgm = *ppgm)) {
        ifBLIN_QV3(options->flags) fprintf( stderr, "pgm %s\n", pgm ? "OK" : "NULL");
        if  (strncmp(pgm->id, "#pGoblin-" VERS, PGOBLIN_STRING_ID_TST)) {
            ifBLIN_QV1(options->flags) warnx("Illegal pgoblin program");
            ERROUT(EX_DATAERR, EINVAL);
        }
    } else {
        if  (!(pgm = malloc(sizeof(pgoblin_tuple) * PGOBLIN_PROGSIZE + sizeof(pgoblin_prog)))) {
            ifBLIN_QV1(options->flags) warnx("Malloc failed in parser #1");
            ERROUT(EX_UNAVAILABLE, ENOMEM);
        }
/*      bzero(pgm, sizeof(pgoblin_tuple) * PGOBLIN_PROGSIZE + sizeof(pgoblin_prog)); */
        strncpy(pgm->id, "#pGoblin-" VERS, PGOBLIN_STRING_ID_LEN);
        pgm->maxx = PGOBLIN_PROGSIZE;
        pgm->curr = 0;
        pgm->debug = NULL;
    }
    ifBLIN_QV4(options->flags) fprintf(stderr, "pgm@%d:%d\n", pgm->curr, pgm->maxx);
    ifBLIN_QV5(options->flags) BLIN_STATEHEADER(stderr, pn)
    lfcount = 1;
    qp = &s;
    qq = &qp;
    r_parm = 0;
    r_lex = 0;
    r_cmd = 0;
    rs = -PGOBLIN_BPARG;
    localflags = PGOBLIN_IGNORE;
    for (state = bg, offset = 0, control = 0; state < xx; offset++, state = control & C_state) {
        if  (offset < txlen) {
            s = text[offset];
            p = class[s];
            if  (p < 0) p = Ls;
            if  (s == '\n') lfcount++;
        } else {
            p = L0;
            s = 0;
        }
        control = automa[state][(u_char)p];
        ifBLIN_QV5(options->flags)
            BLIN_STATEBODY(stderr, pn, rs & 0xFF, clasnames[(u_char)p], text, offset, r_lex, control, statenames, state, C_state, (text[offset + BLIN_STATEVAR_i]>=' '))
        if  (control & Cp) r_parm |= (~(u_int64_t)class[s]) << (rs += PGOBLIN_BPARG);
        if  ((control & Cb) && !r_beg) literal = r_beg = &text[offset];
        if  (control & Ce && r_beg) r_end = literal;
        if  (control & Cc) {
            if  (literal != &text[offset]) *literal = text[offset];
            literal++;
        }
        if  (control & Ci) {
            if  (r_lex >= cmds.szt) r_cmd = ~r_lex;
            ifBLIN_QV3(options->flags)
                fprintf(stderr, "[%d]=%02X(%016llX)\n", pgm->curr, r_cmd, r_parm);
            if  (!r_cmd) {
                ex = EX_DATAERR;
                ifBLIN_QV1(options->flags) warnx("Unknown command at line %d", lfcount);
                break;
            } else if (r_cmd > pgoblin.max) {
                ex = EX_SOFTWARE;
                ifBLIN_QV1(options->flags) warnx("Out of command range at line %d", lfcount);
                break;
            }
            ifBLIN_QV4(options->flags)
                fprintf(stderr, "[%08X]\n", pgoblin.a[r_cmd].flags);
            if  (!(~pgoblin.a[r_cmd].flags & PGOBLIN_ILLEGAL)) {
                ex = EX_SOFTWARE;
                ifBLIN_QV1(options->flags) warnx("Illegal command at line %d", lfcount);
                break;
            } else if (pgm->curr > pgm->maxx) {
                ex = EX_SOFTWARE;
                ifBLIN_QV1(options->flags) warnx("Programm too big");
                break;
            } else if (pgoblin.a[r_cmd].flags & PGOBLIN_IGNORE) {
                localflags |= PGOBLIN_IGNORE;
            } else if (!(localflags & PGOBLIN_IGNORE) || (r_cmd != PGOBLIN_cont)) {
                u_int32_t cc;

                cc = 0;
                localflags &= ~PGOBLIN_IGNORE;
                for (i = 0; i < pgoblin.symparam; i++) {
                    if  (pgoblin.a[r_cmd].flags & pgoblin.p[i].sym) {
                        u_int32_t y = pgoblin.binparams[PREG(r_cmd, i)];
                        cc &= ~(PGOBLIN_ARGMASK << y);
                        cc |= ((r_parm & PGOBLIN_ARGMASK) << y);
                    }
                    r_parm >>= PGOBLIN_BPARG;
                }
                if  (r_cmd & PGOBLIN_EXTENDED) {
                    pgm->code[pgm->curr].cmd = PGOBLIN_extended | (cc & ~PGOBLIN_COMMAND);
                    pgm->code[pgm->curr].e.cmd = (r_cmd & PGOBLIN_COMMAND);
                } else {
                    pgm->code[pgm->curr].cmd = (r_cmd & PGOBLIN_COMMAND) | (cc & ~PGOBLIN_COMMAND);
                    pgm->code[pgm->curr].e.lit = NULL;
                    if  (r_beg && r_end) {
                        *r_end = '\0';
                        if  (!(pgoblin.a[r_cmd].flags & PGOBLIN_NOLITER)) {
                            pgm->code[pgm->curr].e.lit = r_beg;
                        }
                        ifBLIN_QV3(options->flags) {
                            if  (pgm->code[pgm->curr].e.lit)
                                fprintf(stderr, ">%s~\n", pgm->code[pgm->curr].e.lit);
                }   }   }
                pgm->curr++;
        }   }
        if  (control & Co) {
            r_lex = 0;
            r_cmd = 0;
            r_parm = 0;
            rs = -PGOBLIN_BPARG;
            r_beg = 0;
            r_end = 0;
        }
        if  (control & Cl) {
            qp = &s;
            r_lex = babolo_testchar(&cmds, r_lex, qq);
        }
        if  (control & Cx) {
            ifBLIN_QV1(options->flags) fprintf(stderr, " \n");
            errno = EINVAL;
            ex = EX_DATAERR;
        }
        if  (control & Cy) {
            errno = EINVAL;
            ex = EX_DATAERR;
    }   }
out:
    *ppgm = pgm;
    ifBLIN_QV3(options->flags) fprintf(stderr, "-pgoblin_parser %d pgm %s\n", ex, pgm ? "OK" : "NULL");
    return(ex);
}
